// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// Header file for everything.

#include <vector>
#include <string>
#include <SDL/SDL.h>
#include <SDL/SDL_ttf.h>

using namespace std;

#define MAXX 128
#define MAXY 128

#define NEVER 2000000000

int ky[4] = {-1, 0, 1, 0};
int kx[4] = {0, 1, 0, -1};

typedef char int8;
typedef short int16;
typedef long int32;
typedef long long int64;

enum PlayerKey {
  keyUp,
  keyRight,
  keyDown,
  keyLeft,
  keyFire,
  keyEarth,
  keyAir,
  keyWater,
  keyJoinGame // fake key
  };

enum CellType {
  cFree,
  cBreakWall,
  cPermaWall,

  cSnakeHead,
  cSnakeDead,
  cSnakeLeft,
  cSnakeRight,
  cSnakeUp,
  cSnakeDown,
  cSnakeTRON,
  
  cTetris,
  cTetrisDown,
  
  cBig,
  
  cCannonF,
  cCannonA,

  cBonusLevel,
  cBonusScore,
  cBonusAmmo,
  
  cCastleWall,

  cBomb1,
  cBomb2,
  cExplosion
  };

struct TBuffer;

struct TPlayer {
  int8 index;
  int8 typenum;
  int16 lives, frags;
  int32 score, level, ammo, typeval;
  int16 spx, spy; // spawn location
  char *getTypeName();
  int32 color; // color used
  bool key[8]; // keys pressed
  void reactToKey(PlayerKey pk, bool pressed);
  void draw();
  void localDraw() {}
  void init(int c);
  bool useAmmo(int qty) 
    {if(ammo < qty) return false; ammo -= qty; return true;}
  void save(TBuffer& b);
  void load(TBuffer& b);
  };

struct TPlayObject {
  public:
  int8 owner;
  int32 time;
  int getTime() {return time;}
  virtual ~TPlayObject () {}
  virtual void reactToKey(PlayerKey pk, bool pressed) {}
  virtual void draw() {}
  virtual void go() {}
  TPlayer *ow();

  int spd() { return 7999 / (200 + ow()->level) + 1; }
  int scoreGain() { return (100+ow()->level)*(100+ow()->level)/1000; }
  int bulletSpeed() { return 50 + ow()->level/4; }
  virtual void save(TBuffer& b) = 0;
  
  // these methods are supposed to check whether the two collide
  virtual bool reactToBullet(struct OBullet* bullet) {return false;}
  // return true if the bullet is to disappear
  virtual bool reactToPad(struct OPad* pad) {return false;}
  // return true if it kills the pad (reactToPad destroys pad object)
  virtual bool reactToThruster(struct OThruster* bullet) {return false;}
  // return true if it kills the thruster (react destroy thuster)
  
  void killedBy(int player);
  // player killed me -- this deletes the object and scores a frag
  };

struct TPlayground {
  int32 time;
  int32 kqpos;
  
  int64 seed;
  int rand();

  int16 sizeX, sizeY;
  CellType map[MAXY][MAXX];
  char owner[MAXY][MAXX];
  
  void mapSet(int x, int y, CellType c, char o) {
    map[y][x] = c;
    owner[y][x] = o;
    }    
    
  vector<TPlayer*> pl;
  vector<TPlayObject*> objects;

  void upTime(int newTime);
  
  void cleanup();
  void init(int sx, int sy); 
  
  int addPlayer(TPlayer* p);
  
  void TPlayground::deleteObject(TPlayObject *obj);

  void save(TBuffer& b);
  void load(TBuffer& b);
  };

struct TKeypress {
  int32 time;
  bool pressed;
  int32 player; // int32 because also used for color
  PlayerKey key;
  };

template<class T> static inline int size(vector<T> v) {return v.size();}

enum TGameType {gLocal, gHost, gClient};

struct TConfig {
  TGameType gt;
  string server;
  int  port;
  int  res;
  int  sizeX;
  int  sizeY;
  int  resX;
  int  resY;
  bool fs;
  bool centered;

  string err;
  bool ok;
  };

struct TLocalPlayer {
  int index;   // index in the player table
  int keys[8]; // keys for this player
  int color;
  };

struct TIcon {
  int size;
  char pixel[64][64];
  void loadFrom(FILE *f);
  void fillWith(char *chars, int size);
  void drawAt(int x, int y);
  void drawAtCentered(int x, int y) { drawAt(x-size/2, y-size/2); }
  void drawAtRotated(int x, int y, int rx, int ry);
  };

struct TDisplay {

  TIcon permawall;
  TIcon breakwall;
  TIcon freecell;
  TIcon bonusScore;
  TIcon bonusLevel;
  TIcon bonusAmmo;
  TIcon spawncursor;
  TIcon thruster;
  TIcon bullet;
  TIcon ball;
  TIcon bomber;
  TIcon bomb1;
  TIcon bomb2;
  TIcon explosion;
  TIcon castle;
  TIcon cannon;
  TIcon cannona;
  TIcon cannonp;
  TIcon target;
  
  SDL_Surface *surface;
  TTF_Font *font;
  int sizeX;
  int sizeY;
  int ofsX;
  int ofsY;
  int Res;
  bool stop;
  
  int pixelcolor[128];
  
  int delta; // time difference
  
  int getCurrentTime();
  void setCurrentTime(int val);

  void init();
  
  void initForGame();
  
  void loop();
  
  void draw();
  void setIconColor(int);
  
  void putpix(int x, int y, int c);
  
  void addKeypress(int key, bool type);
  
  void putText(int x, int y, int color, const char *text, int align);
  
  void menuLoop();
  void menuKey(int key);
  void menuDraw();
  
  int menupos;
  };

struct TException {
  string msg;
  int data;
  string ds;
  TException(string m, int d=0, string s="") {msg=m; data=d; ds=s;}
  };

#define KQSIZE 32768

TConfig c;
TDisplay d;
TPlayground g;
vector<TLocalPlayer> lps;

TKeypress kqueue[KQSIZE];
int kqend;

void addKeyToQueue(TKeypress& kp);

void startGame();
void dataTransfer();
void broadcastKey(TKeypress& kp);

